#!/usr/bin/env python3

import zarr
import healpy
import numpy as np
from argparse import ArgumentParser


def create_hierachy(dataset, order, startdate, enddate, dt,
                    prefix="healpix_"):
    for o in range(order+1):
        zg = dataset.create_group(f"{prefix}{o}")
        create_dataset(zg, o, startdate, enddate, dt)
        if o < order:
            zg.attrs["hiopy::parent"] = f"{prefix}{o+1}"


def create_dataset(dataset, order, startdate, enddate, dt):
    crs = dataset.create_dataset("crs", data=np.array([np.nan], dtype=np.float32))
    crs.attrs["_ARRAY_DIMENSIONS"] = "crs",
    crs.attrs["grid_mapping_name"] = "healpix"
    crs.attrs["healpix_nside"] = 2**order
    crs.attrs["healpix_order"] = "nest"

    time_data = (np.arange(startdate, enddate, np.timedelta64(dt, 's'))-startdate)//np.timedelta64(1, 's')
    time = dataset.create_dataset("time", data=time_data, fill_value=None)
    time.attrs["_ARRAY_DIMENSIONS"] = ("time",)
    time.attrs["axis"] = "T"
    time.attrs["calendar"] = "proleptic_gregorian"
    time.attrs["units"] = f"seconds since {startdate}"


def _collect_groups(dataset):
    if "crs" in dataset.array_keys():
        yield dataset
    for name, g in dataset.groups():
        yield from _collect_groups(g)


def add_height(dataset, name, n):
    for g in _collect_groups(dataset):
        height = g.create_dataset(name, data=np.arange(n))
        height.attrs["_ARRAY_DIMENSIONS"] = [name]
        height.attrs["axis"] = "Z"
        height.attrs["long_name"] = "generalized_height"
        height.attrs["positive"] = "down"
        height.attrs["standard_name"] = "height"


def add_variable(dataset, name, yac_source_comp, yac_source_grid,
                 zaxis=None, time_method=None,
                 chunk_shape=None, frac_mask=None,
                 yac_name=None):
    for g in _collect_groups(dataset):
        ntime = g.time.shape[0]
        ncells = healpy.nside2npix(g.crs.attrs["healpix_nside"])
        if zaxis is None:
            shape = (ntime, ncells)
            _chunk_shape = np.minimum(chunk_shape, shape) if chunk_shape is not None else None
            v = g.create_dataset(name, shape=shape,
                                 dtype=np.float32, fill_value=np.nan,
                                 chunks=_chunk_shape)
            v.attrs["_ARRAY_DIMENSIONS"] = ("time", "cell")
        else:
            nheight = g[zaxis].shape[0]
            shape = (ntime, nheight, ncells)
            _chunk_shape = np.minimum(chunk_shape, shape) if chunk_shape is not None else None
            v = g.create_dataset(name, shape=shape,
                                 dtype=np.float32, fill_value=np.nan,
                                 chunks=_chunk_shape)
            v.attrs["_ARRAY_DIMENSIONS"] = ("time", zaxis, "cell")
        if time_method is not None:
            v.attrs["hiopy::time_method"] = time_method
        if yac_name is not None:
            v.attrs["hiopy::src_method"] = time_method
        if "hiopy::parent" not in g.attrs:
            v.attrs["hiopy::yac_source"] = (yac_source_comp, yac_source_grid)
        else:
            v.attrs["hiopy::nnn"] = 4


# TODO
# def extent_time()


def consolidate(dataset):
    zarr.convenience.consolidate_metadata(dataset.store)


def info(dataset, tree):
    print(dataset.info)
    if tree:
        print(dataset.tree())


def main():
    parser = ArgumentParser()
    parser.add_argument("dataset", type=zarr.open)
    subparsers = parser.add_subparsers()

    create_hierahchy_parser = subparsers.add_parser("create-hierarchy")
    create_hierahchy_parser.set_defaults(func=create_hierachy)
    create_hierahchy_parser.add_argument("order", type=int)
    create_hierahchy_parser.add_argument("startdate", type=np.datetime64)
    create_hierahchy_parser.add_argument("enddate", type=np.datetime64)
    create_hierahchy_parser.add_argument("dt", type=int)
    create_hierahchy_parser.add_argument("--prefix", type=str, default="healpix_")

    create_dataset_parser = subparsers.add_parser("create-dataset")
    create_dataset_parser.set_defaults(func=create_dataset)
    create_dataset_parser.add_argument("order", type=int)
    create_dataset_parser.add_argument("startdate", type=np.datetime64)
    create_dataset_parser.add_argument("enddate", type=np.datetime64)
    create_dataset_parser.add_argument("dt", type=int)

    add_height_parser = subparsers.add_parser("add-height")
    add_height_parser.set_defaults(func=add_height)
    add_height_parser.add_argument("name", type=str)
    add_height_parser.add_argument("n", type=int)

    add_variable_parser = subparsers.add_parser("add-variable")
    add_variable_parser.set_defaults(func=add_variable)
    add_variable_parser.add_argument("name", type=str)
    add_variable_parser.add_argument("yac_source_comp", type=str)
    add_variable_parser.add_argument("yac_source_grid", type=str)
    add_variable_parser.add_argument("--zaxis", type=str)
    add_variable_parser.add_argument("--time-method", type=str, default="point",
                                     choices=["mean", "point", "sum"])
    add_variable_parser.add_argument("--chunk-shape", type=int, default=None, nargs="*",
                                     help="shape of the data chunks (time, zaxis(if applicable), cell)")
    add_variable_parser.add_argument("--frac-mask", type=str, default=None,
                                     help="Name of the frac_mask array in the same group")
    add_variable_parser.add_argument("--yac-name", type=str, help="name of the yac field")

    consolidate_parser = subparsers.add_parser("consolidate")
    consolidate_parser.set_defaults(func=consolidate)

    info_parser = subparsers.add_parser("info")
    info_parser.set_defaults(func=info)
    info_parser.add_argument("--tree", action="store_true")

    args = parser.parse_args()

    def _main(func, **kwargs):
        return func(**kwargs)
    _main(**vars(args))


if __name__ == "__main__":
    main()
