#!/usr/bin/env python3

from coyote import Coyote, init, run, group_comm_rank, group_comm_size, start_datetime
import zarr
import numpy as np
from argparse import ArgumentParser
import logging
from itertools import chain, groupby
from _distribute_work import distribute_work
from _data_handler import DataHandler
from loco import expose_store_over_http, write_nginx_config

parser = ArgumentParser()
# zarr.copy_store (used with loco) does not work with zarr.open_consolidated
parser.add_argument("datasets", type=zarr.open, nargs="+",
                    help="URL of the dataset. To be passed to zarr.open_consolidated.")
parser.add_argument("--log-level", default=logging.INFO, type=lambda x: getattr(logging, x),
                    help="Configure the logging level.")
parser.add_argument("--process-group", default="hiopy",
                    help="A string determining the process group for distributing the work")
parser.add_argument("--loco", action="store_true",
                    help="Run in LOCO mode")
parser.add_argument("--loco-host", type=str, default="0.0.0.0",
                    help="Host address of the LOCO webserver")
parser.add_argument("--loco-port", type=int, default=8080,
                    help="Port of the LOCO webserver")
parser.add_argument("--loco-capacity", type=int, default=5,
                    help="Number of timesteps to be stored per variable when in LOCO mode")
args = parser.parse_args()

if args.loco:
    assert len(args.datasets) == 1, "Loco only supports reading from 1 dataset"
    loco_store = zarr.MemoryStore()
    zarr.copy_store(args.datasets[0].store, loco_store)
    zarr.convenience.consolidate_metadata(loco_store)
    expose_store_over_http(loco_store, args.loco_host, args.loco_port)
    args.datasets = [zarr.open(store=loco_store)]

init(group_name=args.process_group)
rank, size = group_comm_rank(), group_comm_size()

logging.basicConfig(level=args.log_level,
                    format=f"{args.process_group}({rank}/{size}): %(message)s")


def collect_data_vars(group):
    for name, item in group.arrays():
        if "_ARRAY_DIMENSIONS" in item.attrs:
            if "cell" in item.attrs["_ARRAY_DIMENSIONS"]:
                item.group = group  # amend the variable object with the group
                yield item
    for name, item in group.groups():
        yield from collect_data_vars(item)


def grid_selector(v):
    return v.group.path.split("/")[-1]


data_vars = sorted(chain(*[collect_data_vars(z) for z in args.datasets]), key=grid_selector)
logging.info(f"Found {len(data_vars)} variables")

grouped_data_vars = {grid: list(variables)
                     for grid, variables in groupby(data_vars, key=grid_selector)
                     }
distributed_data_vars = distribute_work(grouped_data_vars,
                                        group_comm_size())

if args.loco:
    write_nginx_config(distributed_data_vars, args.loco_port, group_comm_rank())

my_data_vars = distributed_data_vars[group_comm_rank()]
# my_data_vars list of tuples: ((group, slice), list of variables)

if group_comm_rank() == 0:
    logging.debug(distributed_data_vars)


frac_masks = {}  # the numpy array must keep alive until the end of the program

coyote_instances = {(group, vgroup): Coyote(f"{args.process_group}_{group}_{vgroup}")
                    for group, vgroup, chunk_slice, data_vars in my_data_vars}

for group, vgroup, chunk_slice, data_vars in my_data_vars:
    coyote = coyote_instances[(group, vgroup)]
    nside = data_vars[0].group["crs"].attrs["healpix_nside"]
    ncells_global = 12*(nside**2)
    mi, ma, st = chunk_slice.indices(ncells_global)
    assert st == 1, "Invalid chunk slice"
    ncells = ma-mi
    healpix_order = nside.bit_length() - 1
    coyote.def_healpix_grid(healpix_order, (mi, ma))

    data_handlers = []

    for v in data_vars:
        # compute timestep
        assert "seconds since " in v.group.time.attrs["units"], "Currently the time must be given in seconds"
        dt = v.group.time[1]-v.group.time[0]

        # compute time start index
        t0 = (np.datetime64(start_datetime()) - np.datetime64(v.group.time.attrs["units"][len("seconds since "):]))/np.timedelta64(1, 's')
        t0_idx = np.searchsorted(v.group.time, t0)
        assert v.group.time[t0_idx] == t0, "start_datetime not found in time axis"

        # see YAC_REDUCTION_TIME_NONE etc. (TODO: pass constants through coyote)
        time_methods2yac = {"point": 0,
                            "sum":   1,
                            "mean":  2,
                            "min":   3,
                            "max":   4}

        collection_size = v.shape[1] if len(v.shape) == 3 else 1
        src_name = v.attrs.get("hiopy::src_name", default=v.name.split("/")[-1])
        if "hiopy::yac_source" in v.attrs:
            src_comp, src_grid = v.attrs["hiopy::yac_source"]
        else:
            # find the component with that variables
            src_group, src_vgroup = [(g, vg) for g, vg, _, gvars in chain(*distributed_data_vars)
                                     if g == v.group.attrs["hiopy::parent"] and
                                     len([_ for v in gvars if v.name.split("/")[-1] == src_name]) > 0][0]
            src_comp = src_grid = f"{args.process_group}_{src_group}_{src_vgroup}"
        time_method = v.attrs.get("hiopy::time_method", "point")
        nnn = v.attrs.get("hiopy::nnn", 1)
        frac_mask_name = v.attrs.get("hiopy::frac_mask", None)
        frac_mask = []
        if frac_mask_name is not None:
            if frac_mask_name not in frac_masks[nside]:
                frac_masks[nside][frac_mask_name] = np.array(v.group[frac_mask_name][chunk_slice])
            frac_mask = frac_masks[nside][frac_mask_name]

        logging.info(f"registering {v.name} with {collection_size=}, {src_comp=}, {src_grid=}, {src_name=}, {time_method=}")
        data_handlers.append(DataHandler(v, t0_idx, chunk_slice,
                                         prune_offset=args.loco_capacity if args.loco else None),
                             )
        coyote.add_field(src_comp,
                         src_grid,
                         src_name,
                         data_handlers[-1],
                         f"PT{dt}S",
                         collection_size,
                         yac_time_reduction=time_methods2yac[time_method],
                         interpolation_stack_yaml=f"- nnn:\n    n: {nnn}",
                         frac_mask=frac_mask
                         )

run()

for dh in data_handlers:
    dh.flush()
