#include <pybind11/pybind11.h>
#include <pybind11/stl.h>
#include <pybind11/functional.h>
#include <pybind11/numpy.h>

#include <coyote.hpp>

#define STRINGIFY(x) #x
#define MACRO_STRINGIFY(x) STRINGIFY(x)

namespace py = pybind11;
using namespace pybind11::literals;

using namespace coyote;

namespace {
  template<class T, int flags>
  std::span<const T> _array2span(py::array_t<T, flags> x){
    return std::span<const T>(x.data(), x.size());
  }
}

PYBIND11_MODULE(coyote, m) {
  py::class_<Coyote>(m, "Coyote")
    .def(py::init<std::string>(),
         "name"_a)
    // we need to wrap add_fields to turn the frac_mask buffer into a double*
    .def("add_field", &Coyote::add_field,
         "Add a field to the coyote instance",
         "source_comp"_a,
         "source_grid"_a,
         "field_name"_a,
         "eventHandler"_a,
         "iso_dt"_a = std::nullopt,
         "collection_size"_a = std::nullopt,
         "interpolation_stack_yaml"_a = "[nnn]",
         py::arg("yac_time_reduction") = 0,
         py::arg("frac_mask") = std::vector<double>(),
         py::arg("pass_through") = true,
         py::arg("src_lag") = 1
         )
    .def_property_readonly("comp_comm", [](const Coyote& c) -> int {
      return MPI_Comm_c2f(c.get_mpi_comm());
    })
    .def_property_readonly("comp_rank", [](const Coyote& c) -> int {
      int rank = -1;
      MPI_Comm_rank(c.get_mpi_comm(), &rank);
      return rank;
    })
    .def_property_readonly("comp_size", [](const Coyote& c) -> int {
      int size = -1;
      MPI_Comm_size(c.get_mpi_comm(), &size);
      return size;
    })
    .def("def_grid",
         [](Coyote& c,
           py::array_t<int, py::array::c_style | py::array::forcecast> num_vertices_per_cell,
           py::array_t<double, py::array::c_style | py::array::forcecast> x_vertices,
           py::array_t<double, py::array::c_style | py::array::forcecast> y_vertices,
           py::array_t<int, py::array::c_style | py::array::forcecast> cell_to_vertex,
           std::optional<py::array_t<double, py::array::c_style | py::array::forcecast>> x_cells,
           std::optional<py::array_t<double, py::array::c_style | py::array::forcecast>> y_cells
           ){
          std::optional<std::span<const double>> _x_cells = std::nullopt;
          if(x_cells) _x_cells = _array2span(x_cells.value());
          std::optional<std::span<const double>> _y_cells = std::nullopt;
          if(y_cells) _y_cells = _array2span(y_cells.value());
          c.def_grid(_array2span(num_vertices_per_cell),
                     _array2span(x_vertices),
                     _array2span(y_vertices),
                     _array2span(cell_to_vertex),
                     _x_cells,
                     _y_cells
                     );
        },
        "num_vertices_per_cell"_a,
        "x_vertices"_a,
        "y_vertices"_a,
        "cell_to_vertex"_a,
        py::arg("x_cells") = std::nullopt,
        py::arg("y_cells") = std::nullopt
        )
    .def("def_healpix_grid", &Coyote::def_healpix_grid)
    .def("get_field_collection_size", &Coyote::get_field_collection_size)
    .def("get_field_timestep", &Coyote::get_field_timestep)
    .def("get_field_metadata", &Coyote::get_field_metadata)
    .def("get_all_comp_grid_fields", &Coyote::get_all_comp_grid_fields)
    ;

  py::class_<Event>(m, "Event")
    .def_property_readonly("data", [](const Event& self){
      return py::memoryview::from_buffer(self.data.data(),   // buffer pointer
                                         { (py::ssize_t)self.collection_size, (py::ssize_t)self.point_size },   // shape (rows, cols)
                                         { sizeof(double), sizeof(double)*self.collection_size }  // strides
                                         );
    });

  m.def("run", &coyote_run,
        py::call_guard<py::gil_scoped_release>());
  m.def("init", &coyote_init,
        "group_name"_a);
  m.def("start_datetime", &coyote_start_datetime);
  m.def("end_datetime", &coyote_end_datetime);
  m.def("group_comm", [](){
    return MPI_Comm_c2f(coyote_group_comm());
  });
  m.def("group_comm_rank", [](){
    int r = -1;
    MPI_Comm_rank(coyote_group_comm(), &r);
    return r;
  });
  m.def("group_comm_size", [](){
    int s = 0;
    MPI_Comm_size(coyote_group_comm(), &s);
    return s;
  });

  m.attr("__version__") = MACRO_STRINGIFY(VERSION_INFO);
}
