#!/usr/bin/env python3

from argparse import ArgumentParser, FileType
import numpy as np
import netCDF4
from coyote import Coyote, coyote_def_grid
import logging

parser = ArgumentParser()
parser.add_argument("--component_name", type=str, default="netcdf_output")
parser.add_argument("grid", type=str,
                    help="Grid file (ICON format)")
parser.add_argument("output", type=str,
                    help="Output filename")
parser.add_argument("timestep", type=str,
                    help="Output timestep width in iso format")
parser.add_argument("src_comp", type=str,
                    help="Component name of the yac source component")
parser.add_argument("src_grid", type=str,
                    help="Grid name of the yac source grid")
parser.add_argument("vars", type=str, nargs="+",
                    help="variables names of variables that should be received")

args = parser.parse_args()

coyote = Coyote(args.component_name)

# read grid
grid_ds = netCDF4.Dataset(args.grid, "r")
ncells = grid_ds.dimensions["cell"].size
nverts = grid_ds.dimensions["nv"].size
coyote.def_grid(nverts*np.ones([ncells], dtype=int),
                grid_ds["vlon"][:], grid_ds["vlat"][:],
                grid_ds["vertex_of_cell"][:].T-1,
                grid_ds["clon"][:], grid_ds["clat"][:])

height = coyote.get_field_collection_size(args.src_comp,
                                          args.src_grid,
                                          args.vars[0])

# setup output file
output_ds = netCDF4.Dataset(args.output, "w")
time_dim = output_ds.createDimension("time", None)
cell_dim = output_ds.createDimension("cell", ncells)
height_dim = output_ds.createDimension("height", height)


class handler:
    def __init__(self, var):
        self.var = var
        self.time_counter = 0

    def __call__(self, event):
        self.var[self.time_counter, :, :] = event.data
        logging.info(f"Received {self.var.name} at {self.time_counter}")
        self.time_counter += 1


for vn in args.vars:
    var = output_ds.createVariable(vn, "f4", (time_dim, height_dim, cell_dim))

    coyote.add_field(args.src_comp,
                     args.src_grid,
                     vn,
                     handler(var),
                     args.timestep)

coyote.run()
