#include <memory>
#include <cassert>
#include <numeric>

#include <coyote.hpp>
#include "coyoteenv.hpp"

extern "C" {
#include <yac.h>
}

namespace coyote {

  std::unique_ptr<CoyoteEnv> CoyoteEnv::_instance = nullptr;

  CoyoteEnv::CoyoteEnv(std::string group_name)
    :
    state_def_comps(false),
    state_sync_def(false),
    state_enddef(false)
  {
    yac_cdef_calendar(YAC_PROLEPTIC_GREGORIAN);
    if(group_name != ""){
      std::vector<const char*> group_names = {"yac", group_name.data()};
      std::vector<MPI_Comm> comms(2);
      int is_mpi_initialized = -1;
      MPI_Initialized(&is_mpi_initialized);
      if(!is_mpi_initialized){
        int provided;
        MPI_Init_thread(NULL, NULL, MPI_THREAD_MULTIPLE, &provided);
        assert(provided==MPI_THREAD_MULTIPLE);
        _finalize_mpi = true;
      }
      yac_cmpi_handshake(MPI_COMM_WORLD, 2, group_names.data(), comms.data());
      group_comm = comms[1];
      yac_cinit_comm_instance(comms[0], &yac_instance_id);
    }else{
      yac_cinit_instance(&yac_instance_id);
    }
  }

  CoyoteEnv& CoyoteEnv::getInstance(){
    if(!_instance)
      _instance = std::make_unique<CoyoteEnv>();
    return *_instance;
  }

  int CoyoteEnv::get_yac_instance_id() const{
    return yac_instance_id;
  }

  std::string CoyoteEnv::start_datetime(){
    ensure_def_comps();
    return yac_cget_start_datetime_instance( yac_instance_id );
  }

  std::string CoyoteEnv::end_datetime(){
    ensure_def_comps();
    return yac_cget_end_datetime_instance( yac_instance_id );
  }

  MPI_Comm CoyoteEnv::get_group_comm() const{
    return group_comm;
  }

  void CoyoteEnv::add_field(std::unique_ptr<Field> field){
    _fields.push_back(std::move(field));
  }

  void CoyoteEnv::ensure_def_comps(){
    if(state_def_comps)
      return;
    yac_cdef_comps_instance(yac_instance_id,
                            nullptr,
                            0,
                            nullptr);
    state_def_comps = true;
  }

  void CoyoteEnv::ensure_sync_def(){
    ensure_def_comps();
    if(state_sync_def)
      return;
    yac_csync_def_instance(yac_instance_id);
    state_sync_def = true;
  }

  void CoyoteEnv::ensure_enddef(){
    ensure_sync_def();
    if(state_enddef)
      return;
    yac_cenddef_instance(yac_instance_id);
    state_enddef = true;
  }

  void CoyoteEnv::run(){
    ensure_enddef();
    put_thread = std::thread([&](){
      while(true){
        try{
          auto event = _put_queue.dequeue();
          int field_id = event->field.yac_source_field_id.value();
          int info = -1, ierror = 0;
          int collection_size = yac_cget_collection_size_from_field_id ( field_id );
          if(!event->field.frac_mask.empty()){
            yac_cput_frac_(field_id,
                           collection_size,
                           event->field.buffer.data(),
                           event->field.frac_mask.data(),
                           &info,
                           &ierror);
          }else{
            yac_cput_(field_id,
                      collection_size,
                      event->field.buffer.data(),
                      &info,
                      &ierror);
          }
        }catch(QueueTerminated e){
          break;
        }
      }
    });

    get_thread = std::thread([&](){
      size_t done = 0;
      int ierror = 0;
      while(done < _fields.size()){
        for(auto& field : _fields){
          // skip if done
          if(field->info == YAC_ACTION_OUT_OF_BOUND)
            continue;

          if(!field->get_started){
            // if the consumers are still working on it continue
            if(!field->semaphore.try_acquire())
              continue;
            yac_cget_async_(field->yac_target_field_id, field->collection_size, field->buffer.data(), &(field->info), &ierror);
            field->get_started = true;
          }

          int ready = 0;
          yac_ctest ( field->yac_target_field_id, &ready );
          if(!ready)
            continue;
          field->get_started = false;
          // fire event
          std::shared_ptr<FieldBuffer> e(new FieldBuffer({
                .data = field->buffer,
                .field = *field
              }), [&](auto p){
                field->semaphore.release(); // release the semaphore if all references of the shared_ptr are desctructed
                delete p;
              });
          _main_queue.enqueue(e);
          if(field->yac_source_field_id)
            _put_queue.enqueue(e);

          if(field->info == YAC_ACTION_GET_FOR_RESTART){
            done++;
            field->info = YAC_ACTION_OUT_OF_BOUND;
          }
        }
      }
      _main_queue.end();
      _put_queue.end();
    });

    while(true){
      try{
        auto event = _main_queue.dequeue();
        Event ce = {
          .data = event->data,
          .collection_size = event->field.collection_size,
          .point_size = event->data.size()/(size_t)event->field.collection_size
        };
        event->field.op(ce);
      }catch(QueueTerminated e){
        break;
      }
    }

    // finalization
    get_thread.join();
    put_thread.join();
    _fields.clear();
    yac_cfinalize_instance(yac_instance_id);
    if(_finalize_mpi)
      MPI_Finalize();
  }
}
