#include <cassert>

#include <mpi.h>
#include <coyote.hpp>

#ifdef HAVE_HEALPIX
#include <healpix_cxx/healpix_base.h>
#endif

void coyote_def_healpix_grid(int order,
                             std::optional<std::tuple<size_t, size_t>> cellidx_range,
                             std::vector<double>& vlon,
                             std::vector<double>& vlat,
                             std::vector<int>& nbrVerticesPerCell,
                             std::vector<int>& cellToVertex,
                             std::vector<double>& clon,
                             std::vector<double>& clat,
                             std::vector<int>& global_cell_idx
                            ){
#ifndef HAVE_HEALPIX
  throw std::runtime_error("Healpix not found. Please reconfigure with healpix");
#else
  Healpix_Ordering_Scheme healpixScheme = NEST;
  constexpr auto vertices_per_cell = 4;
  Healpix_Base2 hb(order, healpixScheme);

  auto [min_idx, max_idx] = cellidx_range.value_or(std::make_tuple(0, hb.Npix()));
  size_t nbrCells = max_idx-min_idx;

  clon.resize(nbrCells);
  clat.resize(nbrCells);
  global_cell_idx.resize(nbrCells);

  std::vector<std::tuple<double, double, int>> vertices;
  vertices.reserve(nbrCells * vertices_per_cell);
  std::vector<vec3> corners;
  for (std::size_t i = min_idx; i < max_idx; ++i) {
    {
      pointing p    = hb.pix2ang(static_cast<int>(i));
      std::size_t idx = i-min_idx;
      clon[idx] = p.phi;
      clat[idx] = (M_PI / 2) - p.theta;
      global_cell_idx[idx] = i;
    }
    hb.boundaries(static_cast<int>(i), 1, corners);
    assert(corners.size() == vertices_per_cell && "Unexpected number of vertices");

    for (std::size_t j = 0; j < vertices_per_cell; ++j) {
      pointing p;
      p.from_vec3(corners[j]);
      double lon        = p.phi;
      double lat        = (M_PI / 2) - p.theta;
      std::size_t index = i * vertices_per_cell + j;
      vertices.push_back({lon, lat, index});
    }
  }

  std::sort(vertices.begin(), vertices.end());

  cellToVertex.resize(nbrCells * vertices_per_cell);
  vlon.clear();
  vlon.reserve(nbrCells + (1 << order));
  vlat.clear();
  vlat.reserve(nbrCells + (1 << order));

  for (std::size_t i = 0; i < nbrCells * vertices_per_cell;) {
    double lon = std::get<0>(vertices[i]);
    double lat = std::get<1>(vertices[i]);
    int index  = static_cast<int>(vlon.size());
    vlon.push_back(lon);
    vlat.push_back(lat);

    while (std::get<0>(vertices[i]) == lon && std::get<1>(vertices[i]) == lat) {
      std::size_t cell_index   = static_cast<std::size_t>(std::get<2>(vertices[i])) - min_idx*vertices_per_cell;
      cellToVertex[cell_index] = index;
      ++i;
    }
  }
  nbrVerticesPerCell.clear();
  nbrVerticesPerCell.resize(nbrCells, vertices_per_cell);
#endif
}
